# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from collections.abc import Collection, Iterable, Mapping
from heapq import heappop, heappush
from typing import TypeVar


class CycleError(ValueError):
    """
    Raised when cycles exist in the input graph.

    The second element in the args attribute of instances will contain the
    sequence of nodes in which the cycle lies.
    """


T = TypeVar("T")


def topological_sort(
    items: Iterable[T],
    dependency_graph: Mapping[T, Collection[T]],
) -> Iterable[T]:
    # Tag each item with its input order
    pqueue = list(enumerate(items))
    ordering = {item: ix for ix, item in pqueue}
    seen_since_last_change = 0
    output: set[T] = set()

    # Map blockers to the list of items they block
    blocked_on: dict[T, set[T]] = defaultdict(set)
    blocked: set[T] = set()

    while pqueue:
        if seen_since_last_change == len(pqueue) + len(blocked):
            raise_cycle_error(ordering, pqueue, blocked_on)

        _, n = heappop(pqueue)

        blockers = {d for d in dependency_graph.get(n, []) if d not in output and d in ordering}
        if not blockers:
            seen_since_last_change = 0
            output.add(n)
            if n in blocked:
                blocked.remove(n)
            yield n
            for b in blocked_on.pop(n, []):
                if not any(b in other for other in blocked_on.values()):
                    heappush(pqueue, (ordering[b], b))
        elif n in blocked:
            seen_since_last_change += 1
        else:
            seen_since_last_change = 0
            blocked.add(n)
            for b in blockers:
                blocked_on[b].add(n)
    if blocked_on:
        raise_cycle_error(ordering, pqueue, blocked_on)


def raise_cycle_error(ordering: dict[T, int], pqueue: list[tuple[int, T]], blocked_on: dict[T, set[T]]) -> None:
    bad = next((item for item in blocked_on if item not in ordering), None)
    if bad:
        msg = f"Dependency graph contains a non-existent node {bad!r}"
        raise ValueError(msg)
    unresolved = {n for _, n in pqueue}
    unresolved.update(*blocked_on.values())
    if unresolved:
        msg = f"Dependency graph loop detected among {unresolved!r}"
        raise CycleError(
            msg,
            sorted(unresolved, key=ordering.get),
        )
    msg = "raise_cycle_error called but no unresolved nodes exist"
    raise AssertionError(msg)



from __future__ import annotations

import asyncio
import contextlib
import importlib.metadata
import importlib.util
import logging
import os
import shlex
import subprocess
import sys
import typing as t
from pathlib import Path
from tempfile import NamedTemporaryFile
from textwrap import dedent

import click
import dotenv
import rtoml
import tabulate
import typer
from rich.logging import RichHandler

from pogo_migrate import migrate, sql, yoyo
from pogo_migrate.config import Config, load_config
from pogo_migrate.migration import Migration, topological_sort
from pogo_migrate.util import get_editor, make_file

logger = logging.getLogger(__name__)

VERBOSITY = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}

tempfile_prefix = "_tmp_pogonew"


def load_dotenv() -> None:
    dotenv.load_dotenv(
        dotenv.find_dotenv(usecwd=True),
        override=True,
    )


def setup_logging(verbose: int = 0) -> None:
    """Configure the logging."""
    logging.basicConfig(
        level=VERBOSITY.get(verbose, logging.DEBUG),
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(
                rich_tracebacks=True,
                show_level=False,
                show_path=False,
                show_time=False,
                tracebacks_suppress=[click],
            ),
        ],
    )
    asyncio_logger = logging.getLogger("asyncio")
    asyncio_logger.disabled = True
    root_logger = logging.getLogger("")
    root_logger.setLevel(VERBOSITY.get(verbose, logging.DEBUG))


def _version_callback(*, value: bool) -> None:
    """Get current cli version."""
    if value:
        version = importlib.metadata.version("pogo-migrate")
        typer.echo(f"pogo-migrate {version}")
        raise typer.Exit


def _callback(
    _version: t.Optional[bool] = typer.Option(
        None,
        "-v",
        "--version",
        callback=_version_callback,
        help="Print version and exit.",
    ),
) -> None: ...


app = typer.Typer(name="pogo", callback=_callback)


@app.command("init")
def init(
    migrations_location: str = typer.Option("./migrations", "-m", "--migrations-location"),
    database_env_key: str = typer.Option("POGO_DATABASE", "-d", "--database-env-key"),
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Initiate pogo configuration.

    Create a migrations folder, and inject pogo configuration into
    pyproject.toml.
    """
    setup_logging(verbose)
    pyproject = Path("pyproject.toml")
    if not pyproject.exists():
        pyproject.touch()

    with pyproject.open() as f:
        data = rtoml.load(f)

    if "tool" in data and "pogo" in data["tool"]:
        logger.error("pogo already configured.")
        logger.warning("\n".join(["", "[tool.pogo]"] + [f'{k} = "{v}"' for k, v in data["tool"]["pogo"].items()]))
        raise typer.Exit(code=1)

    cwd = Path.cwd().absolute()
    p = Path(migrations_location).resolve().absolute()

    try:
        loc = p.relative_to(cwd)
    except ValueError as e:
        logger.error("migrations_location is not a child of current location.")
        raise typer.Exit(code=1) from e

    data = {
        "tool": {
            "pogo": {
                "migrations": f"./{loc}",
                "database_env_key": database_env_key,
            },
        },
    }

    config = rtoml.dumps(data, pretty=True)
    logger.error(config)
    if typer.confirm(f"Write configuration to {pyproject.absolute()}"):
        loc.mkdir(exist_ok=True, parents=True)
        with pyproject.open("a") as f:
            f.write("\n")
            f.write(config)


migration_template = dedent(
    '''\
    """
    {message}
    """

    __depends__ = [{depends}]


    async def apply(db):
        ...


    async def rollback(db):
        ...
    ''',
)

migration_sql_template = dedent(
    """\
    -- {message}
    -- depends: {depends}

    -- migrate: apply

    -- migrate: rollback

    """,
)


def retry() -> str:
    choice = ""
    while choice == "":
        choice = typer.prompt("Retry editing? [Ynqh]", default="y")
        if choice == "q":
            raise typer.Exit(code=0)
        if choice in "yn":
            return choice
        if choice == "h":
            logger.error("""
y: reopen the migration file in your editor
n: save the migration as-is, without re-editing
q: quit without saving the migration
h: show this help""")
        choice = ""

    # Shouldn't be reachable
    return ""


def create_with_editor(config: Config, content: str, extension: str, verbose: int) -> Path:
    editor = get_editor(config)
    tmpfile = NamedTemporaryFile(
        mode="w",
        encoding="UTF-8",
        dir=config.migrations,
        prefix=tempfile_prefix,
        suffix=extension,
        delete=False,
    )

    orig_sys_path = sys.path[::]
    try:
        with tmpfile as f:
            f.write(content)
        editor = [part.format(tmpfile.name) for part in shlex.split(editor)]
        if not any(tmpfile.name in part for part in editor):
            editor.append(tmpfile.name)

        mtime = Path(tmpfile.name).lstat().st_mtime
        sys.path.insert(0, config.migrations)
        while True:
            try:
                subprocess.call(editor)  # noqa: S603
            except OSError:
                logger.error("Error: could not open editor!")
            else:
                if Path(tmpfile.name).lstat().st_mtime == mtime:
                    logger.error("Abort: no changes made")
                    raise typer.Exit(code=1)

            try:
                migration = Migration(None, Path(tmpfile.name))
                migration.load()
                message = migration.__doc__
                break
            except Exception:  # noqa: BLE001
                log = logger.exception if verbose else logger.error
                log("Error loading migration.")
                choice = retry()
                if choice == "n":
                    message = ""
                    break

        filename = make_file(config, message, extension)
        Path(tmpfile.name).rename(filename)
        return filename
    finally:
        sys.path = orig_sys_path
        with contextlib.suppress(OSError):
            Path(tmpfile.name).unlink()


@app.command("new")
def new(
    message: str = typer.Option("", "-m", "--message", help="Message describing focus of the migration."),
    *,
    interactive: bool = typer.Option(True, help="Open migration for editing."),  # noqa: FBT003
    sql_: bool = typer.Option(False, "--sql", help="Generate a sql migration."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    if dotenv:
        load_dotenv()
    setup_logging(verbose)
    config = load_config()

    async def new_() -> None:
        migrations = await sql.read_migrations(config.migrations, db=None)
        migrations = list(topological_sort([m.load() for m in migrations]))

        template = migration_sql_template if sql_ else migration_template
        content = template.format(message=message, depends=migrations[-1].id)
        extension = ".sql" if sql_ else ".py"

        if not interactive:
            fp = make_file(config, message, extension)
            with fp.open("w", encoding="UTF-8") as f:
                f.write(content)
            raise typer.Exit(code=0)

        p = create_with_editor(config, content, extension, verbose)
        logger.error("Created file: %s", p)

    asyncio.run(new_())


@app.command("history")
def history(
    database: t.Optional[str] = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    dotenv: bool = typer.Option(False, help="Load environment from .env."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    if dotenv:
        load_dotenv()
    setup_logging(verbose)
    config = load_config()

    async def history_() -> None:
        applied, unapplied = "A", "U"
        connection_string = database or os.environ[config.database_env_key]
        db = await sql.get_connection(connection_string)
        await sql.ensure_pogo_sync(db)
        migrations = await sql.read_migrations(config.migrations, db)
        migrations = topological_sort([m.load() for m in migrations])
        data = (
            (
                applied if m.applied else unapplied,
                m.id,
                "sql" if m.is_sql else "py",
            )
            for m in migrations
        )
        logger.error(tabulate.tabulate(data, headers=("STATUS", "ID", "FORMAT")))

    asyncio.run(history_())


@app.command("apply")
def apply(
    database: t.Optional[str] = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    dotenv: bool = typer.Option(False, help="Load environment from .env."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    if dotenv:
        load_dotenv()

    async def apply_() -> None:
        setup_logging(verbose)
        config = load_config()

        connection_string = database or os.environ[config.database_env_key]
        db = await sql.get_connection(connection_string)

        try:
            await migrate.apply(config, db)
        except Exception:  # noqa: BLE001
            if verbose > 1:
                logger.exception("Error applying migrations.")
            else:
                logger.error("Error applying migrations.")

    asyncio.run(apply_())


@app.command("rollback")
def rollback(
    database: t.Optional[str] = typer.Option(None, "-d", "--database", help="Database connection string."),
    count: int = typer.Option(
        1,
        "-c",
        "--count",
        help="Number of migrations to rollback",
    ),
    *,
    dotenv: bool = typer.Option(False, help="Load environment from .env."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    if dotenv:
        load_dotenv()
    setup_logging(verbose)
    config = load_config()

    async def rollback_() -> None:
        connection_string = database or os.environ[config.database_env_key]
        db = await sql.get_connection(connection_string)

        try:
            await migrate.rollback(config, db, count=count)
        except Exception:  # noqa: BLE001
            if verbose > 1:
                logger.exception("Error rolling back migrations.")
            else:
                logger.error("Error rolling back migrations.")

    asyncio.run(rollback_())


@app.command("mark")
def mark(
    database: t.Optional[str] = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    dotenv: bool = typer.Option(False, help="Load environment from .env."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    if dotenv:
        load_dotenv()
    setup_logging(verbose)
    config = load_config()

    async def _mark() -> None:
        connection_string = database or os.environ[config.database_env_key]
        db = await sql.get_connection(connection_string)

        await sql.ensure_pogo_sync(db)
        migrations = await sql.read_migrations(config.migrations, db)
        migrations = topological_sort([m.load() for m in migrations])

        with db.transaction():
            for migration in migrations:
                migration.load()
                if not migration.applied:
                    if not typer.confirm(f"Mark {migration.id} as applied?"):
                        break

                    await sql.migration_applied(db, migration.id, migration.hash)

    asyncio.run(_mark())


@app.command("unmark")
def unmark(
    database: t.Optional[str] = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    dotenv: bool = typer.Option(False, help="Load environment from .env."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    if dotenv:
        load_dotenv()
    setup_logging(verbose)
    config = load_config()

    async def _unmark() -> None:
        connection_string = database or os.environ[config.database_env_key]
        db = await sql.get_connection(connection_string)

        await sql.ensure_pogo_sync(db)
        migrations = await sql.read_migrations(config.migrations, db)
        migrations = reversed(list(topological_sort([m.load() for m in migrations])))

        with db.transaction():
            try:
                for migration in migrations:
                    migration.load()
                    if migration.applied:
                        if not typer.confirm(f"Unmark {migration.id} as applied?"):
                            break

                        await sql.migration_unapplied(db, migration.id)
            except Exception as e:
                logger.warning(str(e))
                raise

    asyncio.run(_unmark())


@app.command("migrate-yoyo")
def migrate_yoyo(
    database: t.Optional[str] = typer.Option(None, "-d", "--database", help="Database connection string."),
    *,
    dotenv: bool = typer.Option(False, help="Load environment from .env."),  # noqa: FBT003
    verbose: int = typer.Option(
        0,
        "-v",
        "--verbose",
        help="Verbose output. Use multiple times to increase level of verbosity.",
        count=True,
        max=3,
    ),
) -> None:
    """Migrate existing 'yoyo' migrations to 'pogo'."""
    if dotenv:
        load_dotenv()
    setup_logging(verbose)
    config = load_config()

    async def _migrate() -> None:
        for path in config.migrations.iterdir():
            if path.name.endswith(".rollback.sql"):
                continue

            if path.suffix == ".sql":
                content = yoyo.convert_sql_migration(path)

                with path.open("w") as f:
                    f.write(content)
                logger.error("Converted '%s' successfully.", path)
            else:
                logger.error("Python files can not be migrated reliably, please manually update '%s'.", path)

        connection_string = database or os.environ[config.database_env_key]
        db = await sql.get_connection(connection_string)

        await sql.ensure_pogo_sync(db)
        await yoyo.copy_yoyo_migration_history(db)

    asyncio.run(_migrate())



from __future__ import annotations

import logging
import typing as t

import asyncpg

from pogo_migrate.migration import Migration

if t.TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


async def get_connection(connection_string: str) -> asyncpg.Connection:
    return await asyncpg.connect(connection_string)


async def read_migrations(migrations_location: Path, db: asyncpg.Connection | None) -> list[Migration]:
    applied_migrations = await get_applied_migrations(db) if db else []
    return [
        Migration(path.stem, path, applied_migrations)
        for path in migrations_location.iterdir()
        if path.suffix in {".py", ".sql"}
    ]


async def get_applied_migrations(db: asyncpg.Connection) -> set[str]:
    stmt = """
    SELECT
        migration_id
    FROM _pogo_migration
    """
    results = await db.fetch(stmt)

    return {r["migration_id"] for r in results}


async def ensure_pogo_sync(db: asyncpg.Connection) -> None:
    stmt = """
    SELECT exists (
        SELECT FROM pg_tables
        WHERE  schemaname = 'public'
        AND    tablename  = '_pogo_version'
    );
    """
    r = await db.fetchrow(stmt)
    if not r["exists"]:
        stmt = """
        CREATE TABLE _pogo_migration (
            migration_hash VARCHAR(64),  -- sha256 hash of the migration id
            migration_id VARCHAR(255),   -- The migration id (ie path basename without extension)
            applied TIMESTAMPTZ,         -- When this id was applied
            PRIMARY KEY (migration_hash)
        )
        """
        await db.execute(stmt)

        stmt = """
        CREATE TABLE _pogo_version (
            version INT NOT NULL PRIMARY KEY,
            installed TIMESTAMPTZ
        )
        """
        await db.execute(stmt)

        stmt = """
        INSERT INTO _pogo_version (version, installed) VALUES (0, now())
        """
        await db.execute(stmt)


async def migration_applied(db: asyncpg.Connection, migration_id: str, migration_hash: str) -> None:
    stmt = """
    INSERT INTO _pogo_migration (
        migration_hash,
        migration_id,
        applied
    ) VALUES (
        $1, $2, now()
    )
    """
    await db.execute(stmt, migration_hash, migration_id)


async def migration_unapplied(db: asyncpg.Connection, migration_id: str) -> None:
    stmt = """
    DELETE FROM _pogo_migration
    WHERE migration_id = $1
    """
    await db.execute(stmt, migration_id)



from __future__ import annotations

import logging
import typing as t

from pogo_migrate import sql
from pogo_migrate.migration import topological_sort

if t.TYPE_CHECKING:
    import asyncpg

    from pogo_migrate.config import Config

logger = logging.getLogger(__name__)


async def apply(config: Config, db: asyncpg.Connection) -> None:
    await sql.ensure_pogo_sync(db)
    migrations = await sql.read_migrations(config.migrations, db)
    migrations = topological_sort([m.load() for m in migrations])

    async with db.transaction():
        try:
            for migration in migrations:
                migration.load()
                if not migration.applied:
                    logger.error("Applying %s", migration.id)
                    await migration.apply(db)
                    await sql.migration_applied(db, migration.id, migration.hash)
        except Exception as e:
            logger.warning(str(e))
            raise


async def rollback(config: Config, db: asyncpg.Connection, count: int | None = None) -> None:
    await sql.ensure_pogo_sync(db)
    migrations = await sql.read_migrations(config.migrations, db)
    migrations = reversed(list(topological_sort([m.load() for m in migrations])))

    async with db.transaction():
        i = 0
        try:
            for migration in migrations:
                migration.load()
                if migration.applied and (count is None or i < count):
                    logger.error("Rolling back %s", migration.id)
                    await migration.rollback(db)
                    await sql.migration_unapplied(db, migration.id)
                    i += 1
        except Exception as e:
            logger.warning(str(e))
            raise



import re
from pathlib import Path

import asyncpg


def convert_sql_migration(migration: Path) -> str:
    with migration.open() as f:
        apply_content = f.read()

    rollback = migration.with_suffix(".rollback.sql")
    rollback_content = ""
    if rollback.exists():
        with rollback.open() as f:
            rollback_content = f.read()
        rollback.unlink()

    m = re.match(r".*-- (.*)\s-- depends:(.*)\s", apply_content)

    message, depends = "--", "-- depends:"
    if m:
        message = f"-- {m[1]}"
        depends = f"-- depends: {m[2]}"

    content = [message, depends]

    content.extend(["", "-- migrate: apply", ""])
    apply_content = re.sub(r".*-- (.*)\s-- depends:(.*)\s", "", apply_content)
    content.append(apply_content.strip())

    content.extend(["", "-- migrate: rollback", ""])
    rollback_content = re.sub(r".*-- (.*)\s-- depends:(.*)\s", "", rollback_content)
    content.append(rollback_content.strip())

    return "\n".join(content)


async def copy_yoyo_migration_history(db: asyncpg.Connection) -> None:
    stmt = """
    INSERT INTO _pogo_migration (migration_hash, migration_id, applied)
    SELECT
        migration_hash, migration_id, applied_at_utc AS applied
    FROM _yoyo_migration
    """
    await db.execute(stmt)



import asyncpg

from pogo_migrate import config, migrate


async def apply(db: asyncpg.Connection) -> None:
    c = config.load_config()
    await migrate.apply(c, db)


async def rollback(db: asyncpg.Connection) -> None:
    c = config.load_config()
    await migrate.rollback(c, db)



from __future__ import annotations

import dataclasses
import logging
from pathlib import Path

import rtoml
import typer

logger = logging.getLogger(__name__)


CONFIG_FILENAME = "pyproject.toml"


@dataclasses.dataclass
class Config:
    migrations: Path
    database_env_key: str

    @classmethod
    def from_dict(cls: type[Config], data: dict) -> Config:
        data["migrations"] = Path(data["migrations"])
        return cls(**data)


def find_config() -> Path | None:
    """Find the closest config file in the cwd or a parent directory"""
    d = Path.cwd()
    while d != d.parent:
        path = d / CONFIG_FILENAME
        if path.is_file():
            return path
        d = d.parent
    return None


def load_config() -> Config:
    config = find_config()
    if config is None:
        logger.error("No configuration found, missing %s, run 'pogo init ...'", CONFIG_FILENAME)
        raise typer.Exit(code=1)

    with config.open() as f:
        data = rtoml.load(f)

    if "tool" not in data or "pogo" not in data["tool"]:
        logger.error("No configuration found, run 'pogo init ...'")
        raise typer.Exit(code=1)

    return Config.from_dict(data["tool"]["pogo"])
